/***********************************************************************
Aggrega: al algorithm for diffusion limited aggregation

Copyright (C) 2007 Andrea Sottoriva

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <mpi.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define WINSIZE 600

#define DEBUG

#ifdef DEBUG
#define dprintf(format, args...) fprintf(stdout, format , ## args)
#else
#define dprintf (void)
#endif

#define SIZE_PARMS 0
#define STR_INIT   1
#define STR_SYNC   2
#define LEFT_SYNC  3
#define RIGHT_SYNC 4
#define DOWN_SYNC  5
#define UP_SYNC    6
#define DATA_GATH  7

#define TRUE  1
#define FALSE 0

float *lattice;			/* main lattice */
int   *lattice_src;		/* sources/sinks lattice */

float *up, *down;
float *right, *left;		/* up/down/left/right borders */

int *src_up, *src_down;		
int *src_right, *src_left;      /* up/down/left/right source borders */

float D, dt, dx;		/* diffusion coefficent, time and space steps */
float omega = 1.2;		/* SOR omega parameter */
float tot_growth;		/* total growth sites concentration */
float eta = 1.f;		/* shape parameter */

float total_t = 0.f;		/* time counter */
int delay = 0;			/* frame delay */

/* dimensions and domain decomposition */
int size;
int nproc_x, nproc_y;
int npoints_x, npoints_y;
int offset_x, offset_y;

/* MPI stuff */
char proc_name[MPI_MAX_PROCESSOR_NAME];
int ret, nproc, length, nsteps, rank = -1;
int display = FALSE;
int timedep = FALSE;
MPI_Status status;
MPI_Comm grid2d;

/* global visualization stuff */
Display  *mydisp;
Window   mywin;
GC       mygc;
XColor   col;
XEvent   myevent;
KeySym   mykey;
XSizeHints myhint;
XPointer  arg;
Colormap     cmap;
int      myscreen;
int      xvmin, yvmin, xvmax, yvmax;
unsigned long myfore, myback;
unsigned long grey[256];  

/* initialize an X11 window */
void initX11(void) 
{
  XColor col;
  int i;

  if((mydisp = XOpenDisplay("")) == NULL) {
    fprintf(stderr, "Cannot open display\n");
    exit(-1);
  }

  cmap = DefaultColormap(mydisp, 0);
  myscreen = DefaultScreen(mydisp);
  myfore = WhitePixel(mydisp, myscreen);
  myback = BlackPixel(mydisp, myscreen);
  xvmin = myhint.x = 10;
  yvmin = myhint.y = 10;
  xvmax = myhint.width = myhint.min_width = myhint.max_width = WINSIZE;
  yvmax = myhint.height = myhint.min_height = myhint.max_height = WINSIZE;
  myhint.flags = USPosition | PSize | PMinSize | PMaxSize;
  mywin = XCreateSimpleWindow(mydisp, DefaultRootWindow(mydisp), myhint.x, myhint.y,
			      myhint.width, myhint.height, 5, myfore, myback);
  XSetStandardProperties(mydisp,mywin,"Aggrega (diffusion limited aggregation) - Andrea Sottoriva",
			 "Aggrega (diffusion limited aggregation) - Andrea Sottoriva", None, None, None, &myhint);
  mygc = XCreateGC(mydisp, mywin, 0, 0);
  XSetBackground(mydisp, mygc, myback);
  XSetForeground(mydisp, mygc, myfore);
  XSelectInput(mydisp, mywin, ButtonPressMask | KeyPressMask | ExposureMask | Button1MotionMask | Button3MotionMask);

  /* build a greyscale palette */
  for(i = 0; i < 256; i++) {
    col.red = col.green = col.blue = i*256;
    XAllocColor(mydisp, cmap, &col);
    grey[i]=col.pixel;
  }

  XMapRaised(mydisp, mywin);
  XClearWindow(mydisp, mywin);
}

/* 
   Abort program
 */
void abortMPI(int err, int rank)
{
  char error_str[MPI_MAX_ERROR_STRING];
  int len;
  int init_flag;

  MPI_Error_string(err, error_str, &len);

  if(rank < 0) {
    fprintf(stderr, "Error from %d: %s (errno = %d)\n", rank, error_str, err);
  }
  else {
    fprintf(stderr, "Global error: %s (errno = %d)\n", error_str, err);
  }

  MPI_Initialized(&init_flag);

  if(init_flag)
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);

  exit(EXIT_FAILURE);
} 

/* 
   Print Diffusore usage
 */
void usage(void)
{
  fprintf(stderr, "\n\nUsage: ./diffusore <lattice_side> <number_of_iterations> [-d D_factor] [-t dt] [-x dx] [-e eta] [-display] [-timedep] [-s frame_delay(us)]\n\n");
  exit(-1);
}

/* 
   X11 Keyevent handler
 */
Bool checkKeyEvent(Display* dpy, XEvent* event, XPointer arg)
{
  return event->type == KeyPress;
}

/* 
   Check if the i,j site is a possible growth candidate (one of its neighbours is already a growth site
 */
int isGrowthSite(int j, int k)
{
  if(lattice_src[j * (rank == 0 ? size : npoints_y) + k] == 0 && 
     ((j == 0 ? src_left[k] : lattice_src[(j - 1) *(rank == 0 ? size : npoints_y) + k]) == 2 ||
      (j == npoints_x - 1 ? src_right[k] : lattice_src[(j + 1) * (rank == 0 ? size : npoints_y) + k]) == 2 ||
      (k == 0 ? src_up[j] : lattice_src[j * (rank == 0 ? size : npoints_y) + k - 1]) == 2 ||
      (k == npoints_y - 1 ? src_down[j] : lattice_src[j * (rank == 0 ? size : npoints_y) + k + 1]) == 2))
    return 1;
  else
    return 0;
}

int main(int argc, char** argv)
{
  int i, j, k, v, w;			/* misc indexes */

  /* init MPI */
  if((ret = MPI_Init(&argc, &argv)))
    abortMPI(ret, rank);

  /* get task rank */
  if((ret = MPI_Comm_rank(MPI_COMM_WORLD, &rank)))
    abortMPI(ret, rank);

  /* get total number of tasks */
  if((ret = MPI_Comm_size(MPI_COMM_WORLD, &nproc)))
    abortMPI(ret, rank);
  
  /* get processor name */
  if(MPI_Get_processor_name(proc_name, &length))
    abortMPI(ret, rank);

  /* decompose the processors along X and Y coordinates */
  nproc_x = (int)(sqrt(nproc));
  nproc_y = (int)(nproc / nproc_x);

  dprintf("Total processors %d = [ %d, %d ]\n", nproc, nproc_x, nproc_y);

  /* command line parsing */
  if(rank == 0) {

    /* default parameters settings (WARNING: for convergence (D * dt)/(dx^2) <= 1/4) */
    D = 0.04f;
    dt = 0.1f;
    dx = 0.2f;
    
    if(argc < 3) {
      usage();
    }
    
    size = atoi(argv[1]);
    nsteps = atoi(argv[2]);
    
    for(i = 3; i < argc; ++i) {
      if(strcmp(argv[i], "-d") == 0) { /* diffusion coefficent */
	if(i == argc - 1) usage();
	else {
	  D = atof(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-t") == 0) { /* time step */
	if(i == argc - 1) usage();
	else {
	  dt = atof(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-x") == 0) { /* space step */
	if(i == argc - 1) usage();
	else {
	  dx = atof(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-s") == 0) { /* delay between visualization frames */
	if(i == argc - 1) usage();
	else {
	  delay = atoi(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-e") == 0) { /* eta growth parameter*/
	if(i == argc - 1) usage();
	else {
	  eta = atof(argv[i + 1]);
	  i++;
	}
      }
      else if(strcmp(argv[i], "-display") == 0) { /* on/off display flag */
	display = TRUE;
      }
      else if(strcmp(argv[i], "-timedep") == 0) { /* Jacoby time dependence on/off flag */
	timedep = TRUE;
      }
    }  
  }

  /* dimensions are inverted cause the domain decomposition algorithm split the data in the row-wise sense and 
     the MPI_Cart_create splits them in column-sense
   */
  int dims[] = { nproc_y, nproc_x };

  /* X is periodic, Y is not, for the reason above they're inverted */
  int periods[] = { FALSE, TRUE };
  
  /* create a 2D cartesian grid topology */
  if((ret = MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, FALSE, &grid2d)))
    abortMPI(ret, rank);

  dprintf("Aggrega on %*s, %d\n", length, proc_name, rank);

  /* get the new task rank referring to the topology */
  if((ret = MPI_Comm_rank(grid2d, &rank)))
    abortMPI(ret, rank);

  if(rank == 0) {

    /* allocate the entire lattice for process 0 */
    lattice = (float*)malloc(size * size * sizeof(float));
    lattice_src = (int*)malloc(size * size * sizeof(int));
    
    dprintf("Initializing processes:\tsize = %d\n\t\t\tsteps = %d\n\t\t\tD = %.2f\n\t\t\tdt = %.2f\n\t\t\tdx = %.2f\n", size, nsteps, D, dt, dx);

    /* lattice initialization [x,y] >= 0 sources/sinks, [x,y] < 0 unknown value points */
    for(i = 0; i < size; ++i) {
      for(j = 0; j < size; ++j) {
	if(j < 2 || (j < 10 && i > size / 3 && i < size * 2/3) /*|| (i > (size * 2/3 - 3) && i < (size * 2/3 + 5) && j > (size * 2/3 - 5) && j < (size * 2/3 + 5))*/) {
	  lattice[i * size + j] = 1.f;
	  lattice_src[i * size + j] = 1;
	}
	else if(j > size - 3 && i == size / 3) {
	  lattice[i * size + j] = 0.f;
	  lattice_src[i * size + j] = 2;
	}
/*  	else if(j > size - 3 || (j > (size/3 - 12) && j < (size/3 + 12) && i > (size/2 - 12) && i < (size/2 + 12))) {  */
/*  	  lattice[i * size + j] = 0.f;  */
/*  	  lattice_src[i * size + j] = 1;  */
/*  	}  */
	else {
	  lattice[i * size + j] = 0.f;
	  lattice_src[i * size + j] = 0;
	}
      }
    }

    /* delivery parameters */
    for(i = 1; i < nproc; ++i) {

      int int_params[4];
      float float_params[3];
      float *lattice_tmp;
      int *lattice_src_tmp;

      /* domain decomposition along X */
      npoints_x = (int)(size / nproc_x + (size % nproc_x + nproc_x - (i % nproc_x) - 1) / nproc_x);
      if(npoints_x > (int)(size / nproc_x))
	offset_x = npoints_x * (i % nproc_x);
      else
	offset_x = size - (nproc_x - (i % nproc_x)) * npoints_x;
      
      /* domain decomposition along Y */
      npoints_y = (int)(size / nproc_y + (size % nproc_y + nproc_y - (int)(i / nproc_x) - 1) / nproc_y);
      if(npoints_y > (int)(size / nproc_y))
	offset_y = npoints_y * (int)(i / nproc_x);
      else
	offset_y = size - (nproc_y - (int)(i / nproc_x)) * npoints_y;
      
      dprintf("Deliverying params [ npoints = (%d, %d), offset = (%d, %d) ] to process %d\n", npoints_x, npoints_y, offset_x, offset_y, i);
      
      int_params[0] = npoints_x;
      int_params[1] = npoints_y;
      int_params[2] = nsteps;
      int_params[3] = display;
      float_params[0] = D;
      float_params[1] = dt;
      float_params[2] = dx;
      
      /* send parameters */
      if((ret = MPI_Send(int_params, 4, MPI_INT, i, SIZE_PARMS, grid2d)))
	abortMPI(ret, rank);
      
      if((ret = MPI_Send(float_params, 3, MPI_FLOAT, i, SIZE_PARMS, grid2d)))
	abortMPI(ret, rank);

      /* allocate the temporary sublattice for the process i */
      lattice_tmp = (float*)malloc(npoints_x * npoints_y * sizeof(float));
      lattice_src_tmp = (int*)malloc(npoints_x * npoints_y * sizeof(int));

      /* copy by columns */
      for(k = 0; k < npoints_x; ++k)
	memcpy(&lattice_tmp[k * npoints_y], &lattice[offset_x * size + (size * k) + offset_y], npoints_y * sizeof(float));

      /* send the field */
      if((ret = MPI_Send(lattice_tmp, npoints_x * npoints_y, MPI_FLOAT, i, STR_INIT, grid2d)))
      	abortMPI(ret, rank);

      /* copy by columns */
      for(k = 0; k < npoints_x; ++k)
	memcpy(&lattice_src_tmp[k * npoints_y], &lattice_src[offset_x * size + (size * k) + offset_y], npoints_y * sizeof(int));

      /* send the field */
      if((ret = MPI_Send(lattice_src_tmp, npoints_x * npoints_y, MPI_INT, i, STR_INIT, grid2d)))
      	abortMPI(ret, rank);      

      free(lattice_tmp);
      free(lattice_src_tmp);
    }

    /* domain decomposition for the root process along X */
    npoints_x = (int)(size / nproc_x + (size % nproc_x + nproc_x - 1) / nproc_x);
    if(npoints_x > (int)(size / nproc_x))
      offset_x = 0;
    else
      offset_x = size - nproc_x * npoints_x;
    
    /* domain decomposition for the root process along Y */
    npoints_y = (int)(size / nproc_y + (size % nproc_y + nproc_y - 1) / nproc_y);
    if(npoints_y > (int)(size / nproc_y))
      offset_y = 0;
    else
      offset_y = size - nproc_y * npoints_y;
    
    dprintf("Root process [ npoints = (%d, %d), offset = (%d, %d) ]\n", npoints_x, npoints_y, offset_x, offset_y);
    

    /* open a window */
    if(display)
      initX11();
  }
  else {

    int int_params[4];
    float float_params[3];

    /* receive parameters */
    if((ret = MPI_Recv(int_params, 4, MPI_INT, 0, SIZE_PARMS, grid2d, &status)))
      abortMPI(ret, rank);

    npoints_x = int_params[0];
    npoints_y = int_params[1];
    nsteps = int_params[2];
    display = int_params[3];

    if((ret = MPI_Recv(float_params, 3, MPI_FLOAT, 0, SIZE_PARMS, grid2d, &status)))
      abortMPI(ret, rank);

    D = float_params[0];
    dt = float_params[1];
    dx = float_params[2];

    lattice = (float*)malloc(npoints_x * npoints_y * sizeof(float));
    lattice_src = (int*)malloc(npoints_x * npoints_y * sizeof(int));

    /* receive my own part of the string */
    if((ret = MPI_Recv(lattice, npoints_x * npoints_y, MPI_FLOAT, 0, STR_INIT, grid2d, &status)))
      abortMPI(ret, rank);

    /* receive my own part of the string */
    if((ret = MPI_Recv(lattice_src, npoints_x * npoints_y, MPI_INT, 0, STR_INIT, grid2d, &status)))
      abortMPI(ret, rank);
  }
  
  /* alloc borders */
  right = (float*)malloc((npoints_y / 2) * sizeof(float));
  left  = (float*)malloc((npoints_y / 2) * sizeof(float));
  up    = (float*)malloc((npoints_x / 2) * sizeof(float));
  down  = (float*)malloc((npoints_x / 2) * sizeof(float));

  src_right = (int*)malloc(npoints_y * sizeof(int));
  src_left  = (int*)malloc(npoints_y * sizeof(int));
  src_up    = (int*)malloc(npoints_x * sizeof(int));
  src_down  = (int*)malloc(npoints_x * sizeof(int));

  /* neighbours ranks */
  int l_rank, r_rank, u_rank, d_rank;

  /* topological stuff */
  int tdims[2], pers[2], coords[2];

  /* get the rank of the horizontal neighbours */
  if((ret = MPI_Cart_shift(grid2d, 1, 1, &l_rank, &r_rank)))
    abortMPI(ret, rank);

  /* get the rank of the vertical neighbours */
  if((ret = MPI_Cart_shift(grid2d, 0, 1, &u_rank, &d_rank)))
    abortMPI(ret, rank);

  dprintf("Process %d: left = %d, right = %d, up = %d, down = %d\n", rank, l_rank, r_rank, u_rank, d_rank);

  /* get cardinal dimensions */
  if((ret = MPI_Cartdim_get(grid2d, tdims)))
    abortMPI(ret, rank);
  
  /* get process coordinates inside the grid */
  if((ret = MPI_Cart_get(grid2d, 2, tdims, pers, coords)))
    abortMPI(ret, rank);

  /* same story...swap the coordinates */
  { 
    int temp = coords[0];
    coords[0] = coords[1];
    coords[1] = temp;
  }

  /* compute the diffusion */ 
  for(i = 0; i < nsteps; ++i) {

    /* visualization exit handler */
    if(display && rank == 0 && XCheckIfEvent(mydisp, &myevent, checkKeyEvent, arg) && myevent.xkey.keycode == 9) {
      XFreeGC(mydisp,mygc);
      XDestroyWindow(mydisp,mywin);
      XCloseDisplay(mydisp);
      
      if((ret = MPI_Finalize()))
	abortMPI(ret, rank);
      
      exit(EXIT_SUCCESS);
    }

    /**************************************************************************************
     * SYNCHRONIZATION BLACK
     *
     **************************************************************************************/
    if(nproc > 1) {

      /* myup and mydown couse to the column-wise allocation have to be handle differently */
      float *myup, *mydown, *myleft, *myright;

      myup    = (float*)malloc((npoints_x / 2) * sizeof(float));
      mydown  = (float*)malloc((npoints_x / 2) * sizeof(float));
      myleft  = (float*)malloc((npoints_y / 2) * sizeof(float));
      myright = (float*)malloc((npoints_y / 2) * sizeof(float));

      for(j = 0; j < npoints_x / 2; ++j) {
	myup[j] = lattice[(j * 2 + 1) * (rank == 0 ? size : npoints_y)];
	mydown[j] = lattice[(j * 2 + (npoints_y % 2)) * (rank == 0 ? size : npoints_y) + npoints_y - 1];
      }

      for(j = 0; j < npoints_y / 2; ++j) {
	myleft[j] = lattice[j * 2 + 1];
	myright[j] = lattice[(npoints_x - 1) * (rank == 0 ? size : npoints_y) + j * 2 + (npoints_x % 2)];
      }

      /* synchronization */
      if(coords[0] % 2 == 0) {

	/* RIGHT-LEFT synchronization (for EVEN-rows processes) */
	if(r_rank != rank && l_rank != rank) {

	  /* RIGHT synchronization */
	  if(r_rank >= 0) {
	    if((ret = MPI_Sendrecv(myright, npoints_y / 2 + 1, MPI_FLOAT, r_rank, LEFT_SYNC,
				   right, npoints_y / 2 + 1, MPI_FLOAT, r_rank, RIGHT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: right synchronizing with %d : ", rank, r_rank);
	  }
	  
	  /* LEFT synchronization */
	  if(l_rank >= 0) {
	    if((ret = MPI_Sendrecv(myleft, npoints_y / 2 + 1, MPI_FLOAT, l_rank, RIGHT_SYNC,
				   left, npoints_y / 2 + 1, MPI_FLOAT, l_rank, LEFT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);
	    
	    //dprintf("%d: left synchronizing with %d : ", rank, l_rank);
	  }
	}
      }
      else {
	
	/* RIGHT-LEFT synchronization (for ODD-rows processes) */
	if(r_rank != rank && l_rank != rank) {

	  /* LEFT synchronization */
	  if(l_rank >= 0) {
	    if((ret = MPI_Sendrecv(myleft, npoints_y / 2 + 1, MPI_FLOAT, l_rank, RIGHT_SYNC,
				   left, npoints_y / 2 + 1, MPI_FLOAT, l_rank, LEFT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: left synchronizing with %d : ", rank, l_rank);
	  }

	  /* RIGHT synchronization */
	  if(r_rank >= 0) {
	    if((ret = MPI_Sendrecv(myright, npoints_y / 2 + 1, MPI_FLOAT, r_rank, LEFT_SYNC,
				   right, npoints_y / 2 + 1, MPI_FLOAT, r_rank, RIGHT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: right synchronizing with %d\n", rank, r_rank);
	  }
	}
      }

      if(coords[1] % 2 == 0) {
	
	/* UP-DOWN synchronization (for EVEN-column processes) */
	if(u_rank != rank && d_rank != rank) {

	  /* UP synchronization */
	  if(u_rank >= 0) {
	    if((ret = MPI_Sendrecv(myup, npoints_x / 2 + 1, MPI_FLOAT, u_rank, DOWN_SYNC,
				   up, npoints_x / 2 + 1, MPI_FLOAT, u_rank, UP_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: up synchronizing with %d\n", rank, u_rank);
	  }
	  
	  /* DOWN synchronization */
	  if(d_rank >= 0) {
	    if((ret = MPI_Sendrecv(mydown, npoints_x / 2 + 1, MPI_FLOAT, d_rank, UP_SYNC,
				   down, npoints_x / 2 + 1, MPI_FLOAT, d_rank, DOWN_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: down synchronizing with %d\n", rank, d_rank);
	  }
	}
      }
      else {

	/* UP-DOWN synchronization (for ODD-column processes) */
	if(d_rank != rank && u_rank != rank) {

	  /* DOWN synchronization */
	  if(d_rank >= 0) {
	    if((ret = MPI_Sendrecv(mydown, npoints_x / 2 + 1, MPI_FLOAT, d_rank, UP_SYNC,
				   down, npoints_x / 2 + 1, MPI_FLOAT, d_rank, DOWN_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: down synchronizing with %d\n", rank, d_rank);
	  }

	  /* UP synchronization */
	  if(u_rank >= 0) {
	    if((ret = MPI_Sendrecv(myup, npoints_x / 2 + 1, MPI_FLOAT, u_rank, DOWN_SYNC,
				   up, npoints_x / 2 + 1, MPI_FLOAT, u_rank, UP_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: up synchronizing with %d\n", rank, u_rank);
	  }
	}
      }
      
      free(myup);
      free(mydown);
      free(myright);
      free(myleft);
    }

    /**************************************************************************************
     * COMPUTATION RED
     *
     **************************************************************************************/
    for(j = 0; j < npoints_x; ++j) {
      for(k = j % 2; k < npoints_y; k += 2) {

	float r, l, u, d;
	
	if(j > 0 || l_rank == rank)
	  l = lattice[((j - 1 + npoints_x) % npoints_x) * (rank == 0 ? size : npoints_y) + k];
	else
	  l = left[k / 2];
	
	if(j < npoints_x - 1 || r_rank == rank)
	  r = lattice[((j + 1) % npoints_x) * (rank == 0 ? size : npoints_y) + k];
	else
	  r = right[k / 2];
	
	if(k > 0 || u_rank == rank)
	  u = lattice[j * (rank == 0 ? size : npoints_y) + ((k - 1 + npoints_y) % npoints_y)];
	else
	  u = up[j / 2];
	
	if(k < npoints_y - 1 || d_rank == rank)
	  d = lattice[j * (rank == 0 ? size : npoints_y) + ((k + 1) % npoints_y)];
	else
	  d = down[j / 2];
	
	if(lattice_src[j * (rank == 0 ? size : npoints_y) + k] == 0)  {
	  lattice[j * (rank == 0 ? size : npoints_y) + k] = (omega / 4) * (u + d + l + r) +
	    (1 - omega) * lattice[j * (rank == 0 ? size : npoints_y) + k];
	}
      }
    }
    


    /**************************************************************************************
     * SYNCHRONIZATION RED
     *
     **************************************************************************************/
    if(nproc > 1) {

      /* myup and mydown couse to the column-wise allocation have to be handle differently */
      float *myup, *mydown, *myleft, *myright;

      myup    = (float*)malloc((npoints_x / 2) * sizeof(float));
      mydown  = (float*)malloc((npoints_x / 2) * sizeof(float));
      myleft  = (float*)malloc((npoints_y / 2) * sizeof(float));
      myright = (float*)malloc((npoints_y / 2) * sizeof(float));

      for(j = 0; j < npoints_x / 2; ++j) {
	myup[j] = lattice[j * 2 * (rank == 0 ? size : npoints_y)];
	mydown[j] = lattice[(j * 2 + 1 - (npoints_y % 2)) * (rank == 0 ? size : npoints_y) + npoints_y - 1];
      }

      for(j = 0; j < npoints_y / 2; ++j) {
	myleft[j] = lattice[j * 2];
	myright[j] = lattice[(npoints_x - 1) * (rank == 0 ? size : npoints_y) + j * 2 + 1 - (npoints_x % 2)];
      }

      /* synchronization */
      if(coords[0] % 2 == 0) {

	/* RIGHT-LEFT synchronization (for EVEN-rows processes) */
	if(r_rank != rank && l_rank != rank) {

	  /* RIGHT synchronization */
	  if(r_rank >= 0) {
	    if((ret = MPI_Sendrecv(myright, npoints_y / 2 + 1, MPI_FLOAT, r_rank, LEFT_SYNC,
				   right, npoints_y / 2 + 1, MPI_FLOAT, r_rank, RIGHT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: right synchronizing with %d : ", rank, r_rank);
	  }
	  
	  /* LEFT synchronization */
	  if(l_rank >= 0) {
	    if((ret = MPI_Sendrecv(myleft, npoints_y / 2 + 1, MPI_FLOAT, l_rank, RIGHT_SYNC,
				   left, npoints_y / 2 + 1, MPI_FLOAT, l_rank, LEFT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);
	    
	    //dprintf("%d: left synchronizing with %d : ", rank, l_rank);
	  }
	}
      }
      else {
	
	/* RIGHT-LEFT synchronization (for ODD-rows processes) */
	if(r_rank != rank && l_rank != rank) {

	  /* LEFT synchronization */
	  if(l_rank >= 0) {
	    if((ret = MPI_Sendrecv(myleft, npoints_y / 2 + 1, MPI_FLOAT, l_rank, RIGHT_SYNC,
				   left, npoints_y / 2 + 1, MPI_FLOAT, l_rank, LEFT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: left synchronizing with %d : ", rank, l_rank);
	  }

	  /* RIGHT synchronization */
	  if(r_rank >= 0) {
	    if((ret = MPI_Sendrecv(myright, npoints_y / 2 + 1, MPI_FLOAT, r_rank, LEFT_SYNC,
				   right, npoints_y / 2 + 1, MPI_FLOAT, r_rank, RIGHT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: right synchronizing with %d\n", rank, r_rank);
	  }
	}
      }

      if(coords[1] % 2 == 0) {
	
	/* UP-DOWN synchronization (for EVEN-column processes) */
	if(u_rank != rank && d_rank != rank) {

	  /* UP synchronization */
	  if(u_rank >= 0) {
	    if((ret = MPI_Sendrecv(myup, npoints_x / 2 + 1, MPI_FLOAT, u_rank, DOWN_SYNC,
				   up, npoints_x / 2 + 1, MPI_FLOAT, u_rank, UP_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: up synchronizing with %d\n", rank, u_rank);
	  }
	  
	  /* DOWN synchronization */
	  if(d_rank >= 0) {
	    if((ret = MPI_Sendrecv(mydown, npoints_x / 2 + 1, MPI_FLOAT, d_rank, UP_SYNC,
				   down, npoints_x / 2 + 1, MPI_FLOAT, d_rank, DOWN_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: down synchronizing with %d\n", rank, d_rank);
	  }
	}
      }
      else {

	/* UP-DOWN synchronization (for ODD-column processes) */
	if(d_rank != rank && u_rank != rank) {

	  /* DOWN synchronization */
	  if(d_rank >= 0) {
	    if((ret = MPI_Sendrecv(mydown, npoints_x / 2 + 1, MPI_FLOAT, d_rank, UP_SYNC,
				   down, npoints_x / 2 + 1, MPI_FLOAT, d_rank, DOWN_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: down synchronizing with %d\n", rank, d_rank);
	  }

	  /* UP synchronization */
	  if(u_rank >= 0) {
	    if((ret = MPI_Sendrecv(myup, npoints_x / 2 + 1, MPI_FLOAT, u_rank, DOWN_SYNC,
				   up, npoints_x / 2 + 1, MPI_FLOAT, u_rank, UP_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: up synchronizing with %d\n", rank, u_rank);
	  }
	}
      }
      
      free(myup);
      free(mydown);
      free(myright);
      free(myleft);
    }

    /**************************************************************************************
     * COMPUTATION BLACK
     *
     **************************************************************************************/
    for(j = 0; j < npoints_x; ++j) {
      for(k = 1 - (j % 2); k < npoints_y; k += 2) {
	float r, l, u, d;
	
	if(j > 0 || l_rank == rank)
	  l = lattice[((j - 1 + npoints_x) % npoints_x) * (rank == 0 ? size : npoints_y) + k];
	else
	  l = left[k / 2];
	
	if(j < npoints_x - 1 || r_rank == rank)
	  r = lattice[((j + 1) % npoints_x) * (rank == 0 ? size : npoints_y) + k];
	else
	  r = right[k / 2];
	
	if(k > 0 || u_rank == rank)
	  u = lattice[j * (rank == 0 ? size : npoints_y) + ((k - 1 + npoints_y) % npoints_y)];
	else
	  u = up[j / 2];
	
	if(k < npoints_y - 1 || d_rank == rank)
	  d = lattice[j * (rank == 0 ? size : npoints_y) + ((k + 1) % npoints_y)];
	else
	  d = down[j / 2];
	
	if(lattice_src[j * (rank == 0 ? size : npoints_y) + k] == 0)  {
	  lattice[j * (rank == 0 ? size : npoints_y) + k] = (omega / 4) * (u + d + l + r) +
	    (1 - omega) * lattice[j * (rank == 0 ? size : npoints_y) + k];
	}
      }
    }

    /**************************************************************************************
     * SYNCHRONIZATION GROWTH SITES
     *
     **************************************************************************************/

    if(nproc > 1) {
      
      /* myup and mydown couse to the column-wise allocation have to be handle differently */
      int *myup, *mydown;
      
      myup   = (int*)malloc(npoints_x * sizeof(int));
      mydown = (int*)malloc(npoints_x * sizeof(int));
      
      for(j = 0; j < npoints_x; ++j) {
	myup[j] = lattice_src[j * (rank == 0 ? size : npoints_y)];
	mydown[j] = lattice_src[j * (rank == 0 ? size : npoints_y) + npoints_y - 1];
      }

      //MPI_Barrier(grid2d);

      /* growth sites synchronization */ 
      if(coords[0] % 2 == 0) {

	/* RIGHT-LEFT growth sites synchronization (for EVEN-rows processes) */
	if(r_rank != rank && l_rank != rank) {

	  /* RIGHT growth sites synchronization */
	  if(r_rank >= 0) {
	    if((ret = MPI_Sendrecv(&lattice_src[(npoints_x - 1) * (rank == 0 ? size : npoints_y)], npoints_y, MPI_INT, r_rank, LEFT_SYNC,
				   src_right, npoints_y, MPI_INT, r_rank, RIGHT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: right synchronizing with %d : ", rank, r_rank);
	  }
	  
	  /* LEFT growth sites synchronization */
	  if(l_rank >= 0) {
	    if((ret = MPI_Sendrecv(&lattice_src[0], npoints_y, MPI_INT, l_rank, RIGHT_SYNC,
				   src_left, npoints_y, MPI_INT, l_rank, LEFT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);
	    
	    //dprintf("%d: left synchronizing with %d : ", rank, l_rank);
	  }
	}
      }
      else {
	
	/* RIGHT-LEFT growth sites synchronization (for ODD-rows processes) */
	if(r_rank != rank && l_rank != rank) {

	  /* LEFT growth sites synchronization */
	  if(l_rank >= 0) {
	    if((ret = MPI_Sendrecv(&lattice_src[0], npoints_y, MPI_INT, l_rank, RIGHT_SYNC,
				   src_left, npoints_y, MPI_INT, l_rank, LEFT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: left synchronizing with %d : ", rank, l_rank);
	  }

	  /* RIGHT growth sites synchronization */
	  if(r_rank >= 0) {
	    if((ret = MPI_Sendrecv(&lattice_src[(npoints_x - 1) * (rank == 0 ? size : npoints_y)], npoints_y, MPI_INT, r_rank, LEFT_SYNC,
				   src_right, npoints_y, MPI_INT, r_rank, RIGHT_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: right synchronizing with %d\n", rank, r_rank);
	  }
	}
      }

      if(coords[1] % 2 == 0) {
	
	/* UP-DOWN growth sites synchronization (for EVEN-column processes) */
	if(u_rank != rank && d_rank != rank) {

	  /* UP growth sites synchronization */
	  if(u_rank >= 0) {
	    if((ret = MPI_Sendrecv(myup, npoints_x, MPI_INT, u_rank, DOWN_SYNC,
				   src_up, npoints_x, MPI_INT, u_rank, UP_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: up synchronizing with %d\n", rank, u_rank);
	  }
	  
	  /* DOWN growth sites synchronization */
	  if(d_rank >= 0) {
	    if((ret = MPI_Sendrecv(mydown, npoints_x, MPI_INT, d_rank, UP_SYNC,
				   src_down, npoints_x, MPI_INT, d_rank, DOWN_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: down synchronizing with %d\n", rank, d_rank);
	  }
	}
      }
      else {

	/* UP-DOWN growth sites synchronization (for ODD-column processes) */
	if(d_rank != rank && u_rank != rank) {

	  /* DOWN growth sites synchronization */
	  if(d_rank >= 0) {
	    if((ret = MPI_Sendrecv(mydown, npoints_x, MPI_INT, d_rank, UP_SYNC,
				   src_down, npoints_x, MPI_INT, d_rank, DOWN_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: down synchronizing with %d\n", rank, d_rank);
	  }

	  /* UP growth sites synchronization */
	  if(u_rank >= 0) {
	    if((ret = MPI_Sendrecv(myup, npoints_x, MPI_INT, u_rank, DOWN_SYNC,
				   src_up, npoints_x, MPI_INT, u_rank, UP_SYNC, grid2d, &status)))
	      abortMPI(ret, rank);

	    //dprintf("%d: up synchronizing with %d\n", rank, u_rank);
	  }
	}
      }
      
      free(myup);
      free(mydown);
    } 

    /**************************************************************************************
     * PARALLEL GROWTH SIMULATION (not working)
     *
     **************************************************************************************/
       
/*     { */
/*       tot_growth = 0.f; */
/*       for(v = 0; v < npoints_x; ++v) { */
/* 	for(w = 0; w < npoints_y; ++w) { */
/* 	  if(isGrowthSite(v, w)) */
/* 	    tot_growth += lattice[v * (rank == 0 ? size : npoints_y) + w]; */
/* 	} */
/*       } */
      
/*       { */
/* 	float all_growth[nproc]; */
	
/* 	if((ret = MPI_Allgather(&tot_growth, 1, MPI_FLOAT, all_growth, 1, MPI_FLOAT, grid2d))) */
/* 	  abortMPI(ret, rank); */
	
/* 	tot_growth = 0.f; */
/* 	for(k = 0; k < nproc; ++k) */
/* 	  tot_growth += all_growth[k]; */
/*       } */
/*     } */

    /**************************************************************************************
     * GROWTH SIMULATION
     *
     **************************************************************************************/
    
    for(j = 0; j < npoints_x; ++j) {
      for(k = 0; k < npoints_y; ++k) {
	if(isGrowthSite(j, k)) {
	  
	  float p;
	  float gr = 0.f;
	
	  for(v = 0; v < npoints_x; ++v) {
	    for(w = 0; w < npoints_y; ++w) {
	      if(isGrowthSite(v, w))
		gr += lattice[v * (rank == 0 ? size : npoints_y) + w];
	    }
	  }

	  gr = powf(lattice[j * size + k], eta) / powf(gr, eta);

	  //gr = (gr + tot_growth) / 2;
	  p = (float)random() / RAND_MAX;
	  
	  if(gr > p) {
	    lattice[j * (rank == 0 ? size : npoints_y) + k] = 0.f;
	    lattice_src[j * (rank == 0 ? size : npoints_y) + k] = 2;
	  }
	}
      }
    }


    /* collect all the field subspaces for visualization */
    if(display) {

      if(rank == 0) {
	
	/* window display factor */
	float factor = (float)WINSIZE / size; 
	
	/* domain decomposition indexes for each process */
	int cl_npoints_x[nproc], cl_npoints_y[nproc], cl_offset_x[nproc], cl_offset_y[nproc]; 

	float* lattice_tmp;
	int* lattice_src_tmp;
	
	total_t += dt; 

	for(j = 1; j < nproc; ++j) {
	  
	  /* recompute process domain decomposition */
	  cl_npoints_x[j] = (int)(size / nproc_x + (size % nproc_x + nproc_x - (j % nproc_x) - 1) / nproc_x);
	  if(cl_npoints_x[j] > (int)(size / nproc_x))
	    cl_offset_x[j] = cl_npoints_x[j] * (j % nproc_x);
	  else
	    cl_offset_x[j] = size - (nproc_x - (j % nproc_x)) * cl_npoints_x[j];
	  
	  cl_npoints_y[j] = (int)(size / nproc_y + (size % nproc_y + nproc_y - (int)(j / nproc_x) - 1) / nproc_y);
	  if(cl_npoints_y[j] > (int)(size / nproc_y))
	    cl_offset_y[j] = cl_npoints_y[j] * (int)(j / nproc_x);
	  else
	    cl_offset_y[j] = size - (nproc_y - (int)(j / nproc_x)) * cl_npoints_y[j];

	  lattice_tmp = (float*)(malloc(cl_npoints_x[j] * cl_npoints_y[j] * sizeof(float)));
	  lattice_src_tmp = (int*)(malloc(cl_npoints_x[j] * cl_npoints_y[j] * sizeof(int)));
	  
	  //dprintf("Getting the field [ %d, %d ] [ %d, %d ] from %d\n", cl_npoints_x[j], cl_npoints_y[j], cl_offset_x[j], cl_offset_y[j], j);
	  
	  /* receive the computed field and insert it in the main lattice */
	  if((ret = MPI_Recv(lattice_tmp, cl_npoints_x[j] * cl_npoints_y[j], MPI_FLOAT, j, DATA_GATH, grid2d, &status))) 
	    abortMPI(ret, rank); 

	  for(k = 0; k < cl_npoints_x[j]; ++k)
	    memcpy(&lattice[cl_offset_x[j] * size + (size * k) + cl_offset_y[j]], &lattice_tmp[k * cl_npoints_y[j]], cl_npoints_y[j] * sizeof(float));

	  /* receive the computed field and insert it in the main lattice */
	  if((ret = MPI_Recv(lattice_src_tmp, cl_npoints_x[j] * cl_npoints_y[j], MPI_INT, j, DATA_GATH, grid2d, &status))) 
	    abortMPI(ret, rank); 

	  for(k = 0; k < cl_npoints_x[j]; ++k)
	    memcpy(&lattice_src[cl_offset_x[j] * size + (size * k) + cl_offset_y[j]], &lattice_src_tmp[k * cl_npoints_y[j]], cl_npoints_y[j] * sizeof(int));
	  free(lattice_tmp);
	  free(lattice_src_tmp);
	}

	XSetForeground(mydisp, mygc, myfore);

	/* display the lattice as greyscale cells */
	for(j = 0; j < size; ++j) {
	  for(k = 0; k < size; ++k) {
	    if(lattice_src[j * size + k] == 2) 
	      XSetForeground(mydisp, mygc, 128);
	    else
	      XSetForeground(mydisp, mygc, grey[(int)(lattice[j * size + k] * 255)]);
	    XFillRectangle(mydisp, mywin, mygc, j * factor, k * factor, factor, factor);
	  }
	}
	
	/* eventually display the simulation time */
	if(timedep) {
	  XSetForeground(mydisp, mygc, myfore);	
	  char label[128];
	  snprintf(label, 128, "Diffusore: %d processors, time = %.2f s", nproc, total_t); 
	  XDrawImageString(mydisp, mywin, mygc, 2, 15, label, strlen(label)); 
	}
	
	XSetForeground(mydisp, mygc, 128);
	for(j = 1; j < nproc; ++j) {
	  /* processors subfield separators */
	  XDrawLine(mydisp, mywin, mygc, cl_offset_x[j] * factor, cl_offset_y[j] * factor,
		    (cl_offset_x[j] + cl_npoints_x[j]) * factor, cl_offset_y[j] * factor);
	  XDrawLine(mydisp, mywin, mygc, cl_offset_x[j] * factor, cl_offset_y[j] * factor,
		    cl_offset_x[j] * factor, (cl_offset_y[j] + cl_npoints_y[j]) * factor);
	}

	XFlush(mydisp);
	usleep(delay); 
      }
      else {	/* send the field to the root process */

	if((ret = MPI_Send(lattice, npoints_x * npoints_y, MPI_FLOAT, 0, DATA_GATH, grid2d)))
	  abortMPI(ret, rank);

	if((ret = MPI_Send(lattice_src, npoints_x * npoints_y, MPI_INT, 0, DATA_GATH, grid2d)))
	  abortMPI(ret, rank);
      }
    }
  }

  free(lattice);
  free(lattice_src);
  free(right);
  free(left);
  free(up);
  free(down);

  if((ret = MPI_Finalize()))
    abortMPI(ret, rank);

  if(display && rank == 0) {
    XFreeGC(mydisp,mygc);
    XDestroyWindow(mydisp,mywin);
    XCloseDisplay(mydisp);  
  }

  exit(EXIT_SUCCESS);
}   
